#!/usr/bin/env python

""" MultiQC module to parse output from Bismark """

from __future__ import print_function
from collections import OrderedDict
import logging
import re

from multiqc import config
from multiqc.plots import beeswarm, linegraph, bargraph
from multiqc.modules.base_module import BaseMultiqcModule

# Initialise the logger
log = logging.getLogger(__name__)

# Log parsing regexes
regexes = {
    'alignment': {
        'total_reads': r"Sequence(?:s| pairs) analysed in total:\s+(\d+)",
        'aligned_reads': r"Number of(?: paired-end)? alignments with a unique best hit(?: from the different alignments)?:\s+(\d+)",
        'no_alignments': r"Sequence(?:s| pairs) with no alignments under any condition:\s+(\d+)",
        'ambig_reads': r"Sequence(?:s| pairs) did not map uniquely:\s+(\d+)",
        'discarded_reads': r"Sequence(?:s| pairs) which were discarded because genomic sequence could not be extracted:\s+(\d+)",
        'total_c': r"Total number of C's analysed:\s+(\d+)",
        'meth_cpg': r"Total methylated C's in CpG context:\s+(\d+)",
        'meth_chg': r"Total methylated C's in CHG context:\s+(\d+)",
        'meth_chh': r"Total methylated C's in CHH context:\s+(\d+)",
        'unmeth_cpg': r"Total unmethylated C's in CpG context:\s+(\d+)",
        'unmeth_chg': r"Total unmethylated C's in CHG context:\s+(\d+)",
        'unmeth_chh': r"Total unmethylated C's in CHH context:\s+(\d+)",
        'percent_cpg_meth': r"C methylated in CpG context:\s+([\d\.]+)%",
        'percent_chg_meth': r"C methylated in CHG context:\s+([\d\.]+)%",
        'percent_chh_meth': r"C methylated in CHH context:\s+([\d\.]+)%",
        'strand_ot': r"CT(?:\/GA)?\/CT:\s+(\d+)\s+\(\(converted\) top strand\)",
        'strand_ctot': r"GA(?:\/CT)?\/CT:\s+(\d+)\s+\(complementary to \(converted\) top strand\)",
        'strand_ctob': r"GA(?:\/CT)?\/GA:\s+(\d+)\s+\(complementary to \(converted\) bottom strand\)",
        'strand_ob': r"CT(?:\/GA)?\/GA:\s+(\d+)\s+\(\(converted\) bottom strand\)",
        'strand_directional': r"Option '--(directional)' specified \(default mode\): alignments to complementary strands \(CTOT, CTOB\) were ignored \(i.e. not performed\)"
    },
    'dedup': {
        'aligned_reads': r"Total number of alignments analysed in .+:\s+(\d+)",
        'dup_reads': r"Total number duplicated alignments removed:\s+(\d+)",
        'dup_reads_percent': r"Total number duplicated alignments removed:\s+\d+\s+\(([\d\.]+)%\)",
        'dedup_reads': r"Total count of deduplicated leftover sequences:\s+(\d+)",
        'dedup_reads_percent': r"Total count of deduplicated leftover sequences:\s+\d+\s+\(([\d\.]+)% of total\)"
    },
    'methextract': {
        'total_c': r"Total number of C's analysed:\s+(\d+)",
        'meth_cpg': r"Total methylated C's in CpG context:\s+(\d+)",
        'meth_chg': r"Total methylated C's in CHG context:\s+(\d+)",
        'meth_chh': r"Total methylated C's in CHH context:\s+(\d+)",
        'unmeth_cpg': r"Total C to T conversions in CpG context:\s+(\d+)",
        'unmeth_chg': r"Total C to T conversions in CHG context:\s+(\d+)",
        'unmeth_chh': r"Total C to T conversions in CHH context:\s+(\d+)",
        'percent_cpg_meth': r"C methylated in CpG context:\s+([\d\.]+)%",
        'percent_chg_meth': r"C methylated in CHG context:\s+([\d\.]+)%",
        'percent_chh_meth': r"C methylated in CHH context:\s+([\d\.]+)%"
    }
}

class MultiqcModule(BaseMultiqcModule):

    def __init__(self):

        # Initialise the parent object
        super(MultiqcModule, self).__init__(name='Bismark', anchor='bismark',
        href="http://www.bioinformatics.babraham.ac.uk/projects/bismark/",
        info="is a tool to map bisulfite converted sequence reads and determine"\
        " cytosine methylation states.")

        # Set up data structures
        self.bismark_data = {
            'alignment': {},
            'dedup': {},
            'methextract': {},
            'bam2nuc': {}
        }
        self.bismark_mbias_data = {
            'meth': {'CpG_R1' : {}, 'CHG_R1' : {}, 'CHH_R1' : {}, 'CpG_R2' : {}, 'CHG_R2' : {}, 'CHH_R2' : {}},
            'cov': {'CpG_R1' : {}, 'CHG_R1' : {}, 'CHH_R1' : {}, 'CpG_R2' : {}, 'CHG_R2' : {}, 'CHH_R2' : {}}
        }

        # Find and parse bismark alignment reports
        for f in self.find_log_files('bismark/align'):
            parsed_data = self.parse_bismark_report(f['f'], regexes['alignment'])
            if parsed_data is not None:
                # Calculate percent_aligned - doubles as a good check that stuff has worked
                try:
                    parsed_data['percent_aligned'] = (parsed_data['aligned_reads'] / parsed_data['total_reads']) * 100
                except (KeyError, ZeroDivisionError):
                    log.warning('Error calculating percentage for {} - ignoring sample.'.format(f['fn']))
                else:
                    if f['s_name'] in self.bismark_data['alignment']:
                        log.debug("Duplicate alignment sample log found! Overwriting: {}".format(f['s_name']))
                    self.add_data_source(f, section='alignment')
                    self.bismark_data['alignment'][f['s_name']] = parsed_data

        # Find and parse bismark deduplication reports
        for f in self.find_log_files('bismark/dedup'):
            parsed_data = self.parse_bismark_report(f['f'], regexes['dedup'])
            if parsed_data is not None:
                if f['s_name'] in self.bismark_data['dedup']:
                    log.debug("Duplicate deduplication sample log found! Overwriting: {}".format(f['s_name']))
                self.add_data_source(f, section='deduplication')
                self.bismark_data['dedup'][f['s_name']] = parsed_data

        # Find and parse bismark methylation extractor reports
        for f in self.find_log_files('bismark/meth_extract'):
            parsed_data = self.parse_bismark_report(f['f'], regexes['methextract'])
            s_name = f['s_name']
            if parsed_data is not None:
                if s_name in self.bismark_data['methextract']:
                    log.debug("Duplicate methylation extraction sample log found! Overwriting: {}".format(s_name))
                self.add_data_source(f, s_name, section='methylation_extraction')
                self.bismark_data['methextract'][s_name] = parsed_data

        # Find and parse M-bias plot data
        for f in self.find_log_files('bismark/m_bias', filehandles=True):
            self.parse_bismark_mbias(f)
            self.add_data_source(f, section='m_bias')

        # Find and parse bam2nuc reports
        for f in self.find_log_files('bismark/bam2nuc', filehandles=True):
            self.parse_bismark_bam2nuc(f)
            self.add_data_source(f, section='bam2nuc')

        # Filters to strip out ignored sample names
        for k in self.bismark_data:
            self.bismark_data[k] = self.ignore_samples(self.bismark_data[k])
        for k in self.bismark_mbias_data['meth']:
            self.bismark_mbias_data['meth'][k] = self.ignore_samples(self.bismark_mbias_data['meth'][k])
        for k in self.bismark_mbias_data['cov']:
            self.bismark_mbias_data['cov'][k] = self.ignore_samples(self.bismark_mbias_data['cov'][k])

        num_parsed = len(self.bismark_data['alignment'])
        num_parsed += len(self.bismark_data['dedup'])
        num_parsed += len(self.bismark_data['methextract'])
        num_parsed += len(self.bismark_mbias_data['meth']['CpG_R1'])
        num_parsed += len(self.bismark_data['bam2nuc'])
        if num_parsed == 0:
            raise UserWarning

        # Basic Stats Table
        self.bismark_stats_table()

        # Write out to the report
        if len(self.bismark_data['alignment']) > 0:
            self.write_data_file(self.bismark_data['alignment'], 'multiqc_bismark_alignment', sort_cols=True)
            log.info("Found {} alignment reports".format(len(self.bismark_data['alignment'])))
            self.bismark_alignment_chart()

        if len(self.bismark_data['dedup']) > 0:
            self.write_data_file(self.bismark_data['dedup'], 'multiqc_bismark_dedup', sort_cols=True)
            log.info("Found {} dedup reports".format(len(self.bismark_data['dedup'])))
            self.bismark_dedup_chart()

        if len(self.bismark_data['alignment']) > 0:
            self.bismark_strand_chart()

        if len(self.bismark_data['methextract']) > 0:
            self.write_data_file(self.bismark_data['methextract'], 'multiqc_bismark_methextract', sort_cols=True)
            log.info("Found {} methextract reports".format(len(self.bismark_data['methextract'])))
            self.bismark_methlyation_chart()

        if len(self.bismark_mbias_data['meth']['CpG_R1']) > 0:
            self.bismark_mbias_plot()

        if len(self.bismark_data['bam2nuc']) > 0:
            self.write_data_file(self.bismark_data['bam2nuc'], 'multiqc_bismark_bam2nuc', sort_cols=True)
            log.info("Found {} bismark bam2nuc reports".format(len(self.bismark_data['bam2nuc'])))

    def parse_bismark_report(self, report, regexes):
        """ Search a bismark report with a set of regexes """
        parsed_data = {}
        for k, r in regexes.items():
            r_search = re.search(r, report, re.MULTILINE)
            if r_search:
                try:
                    parsed_data[k] = float(r_search.group(1))
                except ValueError:
                    parsed_data[k] = r_search.group(1) # NaN
        if len(parsed_data) == 0: return None
        return parsed_data

    def parse_bismark_mbias(self, f):
        """ Parse the Bismark M-Bias plot data """
        s = f['s_name']
        self.bismark_mbias_data['meth']['CpG_R1'][s] = {}
        self.bismark_mbias_data['meth']['CHG_R1'][s] = {}
        self.bismark_mbias_data['meth']['CHH_R1'][s] = {}
        self.bismark_mbias_data['cov']['CpG_R1'][s] = {}
        self.bismark_mbias_data['cov']['CHG_R1'][s] = {}
        self.bismark_mbias_data['cov']['CHH_R1'][s] = {}
        self.bismark_mbias_data['meth']['CpG_R2'][s] = {}
        self.bismark_mbias_data['meth']['CHG_R2'][s] = {}
        self.bismark_mbias_data['meth']['CHH_R2'][s] = {}
        self.bismark_mbias_data['cov']['CpG_R2'][s] = {}
        self.bismark_mbias_data['cov']['CHG_R2'][s] = {}
        self.bismark_mbias_data['cov']['CHH_R2'][s] = {}
        key = None
        for l in f['f']:
            if 'context' in l:
                if 'CpG' in l:
                    key = 'CpG'
                elif 'CHG' in l:
                    key = 'CHG'
                elif 'CHH' in l:
                    key = 'CHH'
                if '(R1)' in l:
                    key += '_R1'
                elif '(R2)' in l:
                    key += '_R2'
                else:
                    key += '_R1'
            if key is not None:
                sections = l.split()
                try:
                    pos = int(sections[0])
                    self.bismark_mbias_data['meth'][key][s][pos] = float(sections[3])
                    self.bismark_mbias_data['cov'][key][s][pos] = int(sections[4])
                except (IndexError, ValueError):
                    continue

        # Remove empty dicts (eg. R2 for SE data)
        for t in self.bismark_mbias_data:
            for k in self.bismark_mbias_data[t]:
                self.bismark_mbias_data[t][k] = {
                    s_name: self.bismark_mbias_data[t][k][s_name]
                    for s_name in self.bismark_mbias_data[t][k]
                    if len(self.bismark_mbias_data[t][k][s_name]) > 0
                }

    def parse_bismark_bam2nuc(self, f):
        """ Parse reports generated by Bismark bam2nuc """
        if f['s_name'] in self.bismark_data['bam2nuc']:
            log.debug("Duplicate deduplication sample log found! Overwriting: {}".format(f['s_name']))
        self.add_data_source(f, section='bam2nuc')
        self.bismark_data['bam2nuc'][f['s_name']] = dict()

        headers = None
        for l in f['f']:
            sections = l.rstrip().split("\t")
            if headers is None:
                headers = sections
            else:
                for i, h in enumerate(headers):
                    if i == 0:
                        k = sections[0]
                    else:
                        key = "{}_{}".format(k, h.lower().replace(' ','_'))
                        self.bismark_data['bam2nuc'][f['s_name']][key] = sections[i]

    def bismark_stats_table(self):
        """ Take the parsed stats from the Bismark reports and add them to the
        basic stats table at the top of the report """

        headers = {
            'alignment': OrderedDict(),
            'dedup': OrderedDict(),
            'methextract': OrderedDict(),
            'bam2nuc': OrderedDict()
        }
        headers['methextract']['percent_cpg_meth'] = {
            'title': '% mCpG',
            'description': '% Cytosines methylated in CpG context',
            'max': 100,
            'min': 0,
            'suffix': '%',
            'scale': 'Greens'
        }
        headers['methextract']['percent_chg_meth'] = {
            'title': '% mCHG',
            'description': '% Cytosines methylated in CHG context',
            'max': 100,
            'min': 0,
            'suffix': '%',
            'scale': 'Oranges'
        }
        headers['methextract']['percent_chh_meth'] = {
            'title': '% mCHH',
            'description': '% Cytosines methylated in CHH context',
            'max': 100,
            'min': 0,
            'suffix': '%',
            'scale': 'Oranges'
        }
        headers['methextract']['total_c'] = {
            'title': "M C's",
            'description': 'Total number of C\'s analysed, in millions',
            'min': 0,
            'scale': 'Purples',
            'modify': lambda x: x / 1000000
        }
        headers['bam2nuc']['C_coverage'] = {
            'title': 'C Coverage',
            'description': 'Cyotosine Coverage',
            'min': 0,
            'suffix': 'X',
            'scale': 'Greens',
            'format': '{:,.2f}'
        }
        headers['dedup']['dup_reads_percent'] = {
            'title': '% Dups',
            'description': 'Percent Duplicated Alignments',
            'max': 100,
            'min': 0,
            'suffix': '%',
            'scale': 'RdYlGn-rev',
        }
        headers['dedup']['dedup_reads'] = {
            'title': '{} Unique'.format(config.read_count_prefix),
            'description': 'Deduplicated Alignments ({})'.format(config.read_count_desc),
            'min': 0,
            'scale': 'Greens',
            'modify': lambda x: x * config.read_count_multiplier,
            'shared_key': 'read_count',
            'hidden': True
        }
        headers['alignment']['aligned_reads'] = {
            'title': '{} Aligned'.format(config.read_count_prefix),
            'description': 'Total Aligned Sequences ({})'.format(config.read_count_desc),
            'min': 0,
            'scale': 'PuRd',
            'modify': lambda x: x * config.read_count_multiplier,
            'shared_key': 'read_count',
            'hidden': True
        }
        headers['alignment']['percent_aligned'] = {
            'title': '% Aligned',
            'description': 'Percent Aligned Sequences',
            'max': 100,
            'min': 0,
            'suffix': '%',
            'scale': 'YlGn'
        }

        self.general_stats_addcols(self.bismark_data['methextract'], headers['methextract'])
        self.general_stats_addcols(self.bismark_data['bam2nuc'], headers['bam2nuc'])
        self.general_stats_addcols(self.bismark_data['dedup'], headers['dedup'])
        self.general_stats_addcols(self.bismark_data['alignment'], headers['alignment'])


    def bismark_alignment_chart (self):
        """ Make the alignment plot """

        # Specify the order of the different possible categories
        keys = OrderedDict()
        keys['aligned_reads']   = { 'color': '#2f7ed8', 'name': 'Aligned Uniquely' }
        keys['ambig_reads']     = { 'color': '#492970', 'name': 'Aligned Ambiguously' }
        keys['no_alignments']   = { 'color': '#0d233a', 'name': 'Did Not Align' }
        keys['discarded_reads'] = { 'color': '#f28f43', 'name': 'No Genomic Sequence' }

        # Config for the plot
        config = {
            'id': 'bismark_alignment',
            'title': 'Bismark: Alignment Scores',
            'ylab': '# Reads',
            'cpswitch_counts_label': 'Number of Reads'
        }

        self.add_section (
            name = 'Alignment Rates',
            anchor = 'bismark-alignment',
            plot = bargraph.plot(self.bismark_data['alignment'], keys, config)
        )


    def bismark_strand_chart (self):
        """ Make the strand alignment plot """

        # Specify the order of the different possible categories
        keys = OrderedDict()
        keys['strand_ob']   = { 'name': 'Original bottom strand' }
        keys['strand_ctob'] = { 'name': 'Complementary to original bottom strand' }
        keys['strand_ctot'] = { 'name': 'Complementary to original top strand' }
        keys['strand_ot']   = { 'name': 'Original top strand' }

        # See if we have any directional samples
        directional = 0
        d_mode = ''
        for sn in self.bismark_data['alignment'].values():
            if 'strand_directional' in sn.keys():
                directional += 1
        if directional == len(self.bismark_data['alignment']):
            keys.pop('strand_ctob', None)
            keys.pop('strand_ctot', None)
            d_mode = 'All samples were run with <code>--directional</code> mode; alignments to complementary strands (CTOT, CTOB) were ignored.'
        elif directional > 0:
            d_mode = '{} samples were run with <code>--directional</code> mode; alignments to complementary strands (CTOT, CTOB) were ignored.'.format(directional)

        # Config for the plot
        config = {
            'id': 'bismark_strand_alignment',
            'title': 'Bismark: Alignment to Individual Bisulfite Strands',
            'ylab': '% Reads',
            'cpswitch_c_active': False,
            'cpswitch_counts_label': 'Number of Reads'
        }

        self.add_section (
            name = 'Strand Alignment',
            anchor = 'bismark-strands',
            description = d_mode,
            plot = bargraph.plot(self.bismark_data['alignment'], keys, config)
        )


    def bismark_dedup_chart (self):
        """ Make the deduplication plot """

        # Specify the order of the different possible categories
        keys = OrderedDict()
        keys['dedup_reads'] = { 'name': 'Deduplicated reads (remaining)' }
        keys['dup_reads']   = { 'name': 'Duplicate reads (removed)' }

        # Config for the plot
        config = {
            'id': 'bismark_deduplication',
            'title': 'Bismark: Deduplication',
            'ylab': '% Reads',
            'cpswitch_c_active': False,
            'cpswitch_counts_label': 'Number of Reads'
        }

        self.add_section (
            name = 'Deduplication',
            anchor = 'bismark-deduplication',
            plot = bargraph.plot(self.bismark_data['dedup'], keys, config)
        )



    def bismark_methlyation_chart (self):
        """ Make the methylation plot """

        # Config for the plot
        keys = OrderedDict()
        defaults = {
            'max': 100,
            'min': 0,
            'suffix': '%',
            'decimalPlaces': 1
        }
        keys['percent_cpg_meth'] = dict(defaults, **{ 'title': 'Methylated CpG' })
        keys['percent_chg_meth'] = dict(defaults, **{ 'title': 'Methylated CHG' })
        keys['percent_chh_meth'] = dict(defaults, **{ 'title': 'Methylated CHH' })

        self.add_section (
            name = 'Cytosine Methylation',
            anchor = 'bismark-methylation',
            plot = beeswarm.plot(self.bismark_data['methextract'], keys, {'id': 'bismark-methylation-dp'})
        )


    def bismark_mbias_plot (self):
        """ Make the M-Bias plot """

        description = '<p>This plot shows the average percentage methylation and coverage across reads. See the \n\
        <a href="https://rawgit.com/FelixKrueger/Bismark/master/Docs/Bismark_User_Guide.html#m-bias-plot" target="_blank">bismark user guide</a> \n\
        for more information on how these numbers are generated.</p>'

        pconfig = {
            'id': 'bismark_mbias',
            'title': 'Bismark: M-Bias',
            'ylab': '% Methylation',
            'xlab': 'Position (bp)',
            'xDecimals': False,
            'ymax': 100,
            'ymin': 0,
            'tt_label': '<b>{point.x} bp</b>: {point.y:.1f}%',
            'data_labels': [
                {'name': 'CpG R1', 'ylab': '% Methylation', 'ymax': 100},
                {'name': 'CHG R1', 'ylab': '% Methylation', 'ymax': 100},
                {'name': 'CHH R1', 'ylab': '% Methylation', 'ymax': 100}
            ]
        }
        datasets = [
            self.bismark_mbias_data['meth']['CpG_R1'],
            self.bismark_mbias_data['meth']['CHG_R1'],
            self.bismark_mbias_data['meth']['CHH_R1']
        ]

        if len(self.bismark_mbias_data['meth']['CpG_R2']) > 0:
            pconfig['data_labels'].append({'name': 'CpG R2', 'ylab': '% Methylation', 'ymax': 100})
            pconfig['data_labels'].append({'name': 'CHG R2', 'ylab': '% Methylation', 'ymax': 100})
            pconfig['data_labels'].append({'name': 'CHH R2', 'ylab': '% Methylation', 'ymax': 100})
            datasets.append(self.bismark_mbias_data['meth']['CpG_R2'])
            datasets.append(self.bismark_mbias_data['meth']['CHG_R2'])
            datasets.append(self.bismark_mbias_data['meth']['CHH_R2'])

        self.add_section (
            name = 'M-Bias',
            anchor = 'bismark-mbias',
            description = description,
            plot = linegraph.plot(datasets, pconfig)
        )



#!/usr/bin/env python
# coding=utf-8
from __future__ import print_function

import itertools
import re
from collections import defaultdict
from multiqc.modules.base_module import BaseMultiqcModule
from multiqc.plots import bargraph, table
from .utils import make_headers, Metric

# Initialise the logger
import logging
log = logging.getLogger(__name__)


NAMESPACE = 'DRAGEN mapping'


class DragenMappingMetics(BaseMultiqcModule):
    def add_mapping_metrics(self):
        data_by_rg_by_sample = defaultdict(dict)
        data_by_phenotype_by_sample = defaultdict(dict)

        for f in self.find_log_files('dragen/mapping_metrics'):
            data_by_readgroup, data_by_phenotype = parse_mapping_metrics_file(f)

            if f['s_name'] in data_by_rg_by_sample:
                log.debug('Duplicate DRAGEN output prefix found! Overwriting: {}'.format(f['s_name']))
            self.add_data_source(f, section='stats')
            data_by_phenotype_by_sample[f['s_name']].update(data_by_phenotype)

            for rg, data in data_by_readgroup.items():
                if any(rg in d_rg for sn, d_rg in data_by_rg_by_sample.items()):
                    log.debug('Duplicate read group name {} found for output prefix {}! Overwriting'.format(rg, f['s_name']))
            data_by_rg_by_sample[f['s_name']].update(data_by_readgroup)

        # filter to strip out ignored sample names:
        data_by_rg_by_sample = self.ignore_samples(data_by_rg_by_sample)
        data_by_phenotype_by_sample = self.ignore_samples(data_by_phenotype_by_sample)

        # flattening phenotype-sample data by adding a prefix " normal" to the normal samples
        data_by_sample = dict()
        for sn in data_by_phenotype_by_sample:
            for phenotype in data_by_phenotype_by_sample[sn]:
                new_sn = sn
                if phenotype == 'normal':
                    new_sn = sn + '_normal'
                data_by_sample[new_sn] = data_by_phenotype_by_sample[sn][phenotype]

        if not data_by_rg_by_sample and not data_by_phenotype_by_sample:
            return set()
        self.report_mapping_metrics(data_by_sample, data_by_rg_by_sample)
        return data_by_rg_by_sample.keys()

    def report_mapping_metrics(self, data_by_sample, data_by_rg_by_sample):
        # merging all read group data
        data_by_rg = dict()
        for sname in data_by_rg_by_sample:
            for rg, d in data_by_rg_by_sample[sname].items():
                if rg in data_by_rg:
                    rg = rg + ' (' + sname + ')'
                data_by_rg[rg] = d

        # getting all available metric names to determine table headers
        all_metric_names = set()
        for sn, d_by_rg in data_by_rg_by_sample.items():
            for rg, data in d_by_rg.items():
                for m in data.keys():
                    all_metric_names.add(m)
        # and making headers
        genstats_headers, own_tabl_headers = make_headers(all_metric_names, MAPPING_METRICS)

        self.general_stats_addcols(data_by_sample, genstats_headers, namespace=NAMESPACE)

        self.add_section(
            name='Mapping metrics',
            anchor='dragen-mapping-metrics',
            description="""
            Mapping metrics, similar to the metrics computed by the samtools-stats command.
            Shown on per read group level. To see per-sample level metrics, refer to the general
            stats table.
            """,
            plot=table.plot(data_by_rg, own_tabl_headers, pconfig={'namespace': NAMESPACE})
        )

        # Make bargraph plots of mapped, dupped and paired reads
        self.__map_pair_dup_read_chart(data_by_rg)

    def __map_pair_dup_read_chart(self, data_by_sample):
        chart_data = dict()
        for sample_id, data in data_by_sample.items():
            if data['Not properly paired reads (discordant)'] + data['Properly paired reads']\
                    + data['Singleton reads (itself mapped; mate unmapped)']\
                    + data['Unmapped reads'] != data['Total reads in RG']:
                log.warning("sum of unpaired/discordant/proppaired/unmapped reads not matching total, "
                            "skipping mapping/paired percentages plot for: {}".format(sample_id))
                continue
            if data['Number of unique & mapped reads (excl. duplicate marked reads)']\
                    + data['Number of duplicate marked reads']\
                    + data['Unmapped reads'] != data['Total reads in RG']:
                log.warning("sum of unique/duplicate/unmapped reads not matching total, "
                            "skipping mapping/duplicates percentages plot for: {}".format(sample_id))
                continue
            chart_data[sample_id] = data
        self.add_section(
            name='Mapped / paired / duplicated',
            anchor='dragen-mapped-paired-duplicated',
            description='Distribution of reads based on pairing, duplication and mapping.',
            plot=bargraph.plot([chart_data, chart_data], [
            {
                'Number of unique & mapped reads (excl. duplicate marked reads)': {'color': '#437bb1', 'name': 'Unique'},
                'Number of duplicate marked reads':                               {'color': '#f5a742', 'name': 'Duplicated'},
                'Unmapped reads':                                                 {'color': '#b1084c', 'name': 'Unmapped'},
            },
            {
                'Properly paired reads':                          {'color': '#099109', 'name': 'Paired, properly'},
                'Not properly paired reads (discordant)':         {'color': '#c27a0e', 'name': 'Paired, discordant'},
                'Singleton reads (itself mapped; mate unmapped)': {'color': '#912476', 'name': 'Singleton'},
                'Unmapped reads':                                 {'color': '#b1084c', 'name': 'Unmapped'},
            },
            ], {
                'id': 'mapping_dup_percentage_plot',
                'title': 'Dragen: Mapped/paired/duplicated reads per read group',
                'ylab': 'Reads',
                'cpswitch_counts_label': 'Reads',
                'data_labels': [
                    {
                        'name': 'Unique vs duplicated vs unmapped',
                        'ylab': 'Reads',
                        'cpswitch_counts_label': 'Reads',
                    },
                    {
                        'name': 'Paired vs. discordant vs. singleton',
                        'ylab': 'Reads',
                        'cpswitch_counts_label': 'Reads',
                    }
                ]
            })
        )


def parse_mapping_metrics_file(f):
    """
    Mapping and aligning metrics, like the metrics computed by the Samtools Flagstat command, are available
    on an aggregate level (over all input data), and on a per read group level. Unless explicitly stated,
    the metrics units are in reads (ie, not in terms of pairs or alignments).

    T_SRR7890936_50pc.mapping_metrics.csv

    # phenotype-level metrics (tumor or normal):
    TUMOR MAPPING/ALIGNING SUMMARY,,Total input reads,2200000000,100.00
    TUMOR MAPPING/ALIGNING SUMMARY,,Number of duplicate marked reads,433637413,19.71
    TUMOR MAPPING/ALIGNING SUMMARY,,Number of duplicate marked and mate reads removed,NA
    TUMOR MAPPING/ALIGNING SUMMARY,,Number of unique reads (excl. duplicate marked reads),1766362587,80.29
    TUMOR MAPPING/ALIGNING SUMMARY,,Reads with mate sequenced,2200000000,100.00
    TUMOR MAPPING/ALIGNING SUMMARY,,Reads without mate sequenced,0,0.00
    TUMOR MAPPING/ALIGNING SUMMARY,,QC-failed reads,0,0.00
    TUMOR MAPPING/ALIGNING SUMMARY,,Mapped reads,2130883930,96.86
    TUMOR MAPPING/ALIGNING SUMMARY,,Mapped reads R1,1066701794,96.97
    TUMOR MAPPING/ALIGNING SUMMARY,,Mapped reads R2,1064182136,96.74
    TUMOR MAPPING/ALIGNING SUMMARY,,Number of unique & mapped reads (excl. duplicate marked reads),1697246517,77.15
    TUMOR MAPPING/ALIGNING SUMMARY,,Unmapped reads,69116070,3.14
    TUMOR MAPPING/ALIGNING SUMMARY,,Singleton reads (itself mapped; mate unmapped),3917092,0.18
    TUMOR MAPPING/ALIGNING SUMMARY,,Paired reads (itself & mate mapped),2126966838,96.68
    TUMOR MAPPING/ALIGNING SUMMARY,,Properly paired reads,2103060370,95.59
    TUMOR MAPPING/ALIGNING SUMMARY,,Not properly paired reads (discordant),23906468,1.09
    TUMOR MAPPING/ALIGNING SUMMARY,,Paired reads mapped to different chromosomes,17454370,0.82
    TUMOR MAPPING/ALIGNING SUMMARY,,Paired reads mapped to different chromosomes (MAPQ>=10),6463547,0.30
    TUMOR MAPPING/ALIGNING SUMMARY,,Reads with MAPQ [40:inf),2002661377,91.03
    TUMOR MAPPING/ALIGNING SUMMARY,,Reads with MAPQ [30:40),7169392,0.33
    TUMOR MAPPING/ALIGNING SUMMARY,,Reads with MAPQ [20:30),16644390,0.76
    TUMOR MAPPING/ALIGNING SUMMARY,,Reads with MAPQ [10:20),20280057,0.92
    TUMOR MAPPING/ALIGNING SUMMARY,,Reads with MAPQ [ 0:10),84128714,3.82
    TUMOR MAPPING/ALIGNING SUMMARY,,Reads with MAPQ NA (Unmapped reads),69116070,3.14
    TUMOR MAPPING/ALIGNING SUMMARY,,Reads with indel R1,26849051,2.52
    TUMOR MAPPING/ALIGNING SUMMARY,,Reads with indel R2,24810803,2.33
    TUMOR MAPPING/ALIGNING SUMMARY,,Total bases,330000000000
    TUMOR MAPPING/ALIGNING SUMMARY,,Total bases R1,165000000000
    TUMOR MAPPING/ALIGNING SUMMARY,,Total bases R2,165000000000
    TUMOR MAPPING/ALIGNING SUMMARY,,Mapped bases R1,160005269100
    TUMOR MAPPING/ALIGNING SUMMARY,,Mapped bases R2,159627320400
    TUMOR MAPPING/ALIGNING SUMMARY,,Soft-clipped bases R1,1757128997,1.10
    TUMOR MAPPING/ALIGNING SUMMARY,,Soft-clipped bases R2,3208748350,2.01
    TUMOR MAPPING/ALIGNING SUMMARY,,Mismatched bases R1,585802788,0.37
    TUMOR MAPPING/ALIGNING SUMMARY,,Mismatched bases R2,1155805091,0.72
    TUMOR MAPPING/ALIGNING SUMMARY,,Mismatched bases R1 (excl. indels),501394281,0.31
    TUMOR MAPPING/ALIGNING SUMMARY,,Mismatched bases R2 (excl. indels),1073788605,0.67
    TUMOR MAPPING/ALIGNING SUMMARY,,Q30 bases,297564555927,90.17
    TUMOR MAPPING/ALIGNING SUMMARY,,Q30 bases R1,155492239719,94.24
    TUMOR MAPPING/ALIGNING SUMMARY,,Q30 bases R2,142072316208,86.10
    TUMOR MAPPING/ALIGNING SUMMARY,,Q30 bases (excl. dups & clipped bases),246555769158
    TUMOR MAPPING/ALIGNING SUMMARY,,Total alignments,2190085267
    TUMOR MAPPING/ALIGNING SUMMARY,,Secondary alignments,0
    TUMOR MAPPING/ALIGNING SUMMARY,,Supplementary (chimeric) alignments,59201337
    TUMOR MAPPING/ALIGNING SUMMARY,,Estimated read length,150.00
    TUMOR MAPPING/ALIGNING SUMMARY,,Average sequenced coverage over genome,102.83
    TUMOR MAPPING/ALIGNING SUMMARY,,Insert length: mean,383.00
    TUMOR MAPPING/ALIGNING SUMMARY,,Insert length: median,376.00
    TUMOR MAPPING/ALIGNING SUMMARY,,Insert length: standard deviation,85.15
    # general metrics - reporting in the header (except for DRAGEN mapping rate may be different to T and N:
    TUMOR MAPPING/ALIGNING SUMMARY,,Bases in reference genome,3209286105
    TUMOR MAPPING/ALIGNING SUMMARY,,Bases in target bed [% of genome],NA
    TUMOR MAPPING/ALIGNING SUMMARY,,Provided sex chromosome ploidy,NA
    TUMOR MAPPING/ALIGNING SUMMARY,,DRAGEN mapping rate [mil. reads/second],0.39
    # then same for normal:
    NORMAL MAPPING/ALIGNING SUMMARY,,Total input reads,1100000000,100.00
    NORMAL MAPPING/ALIGNING SUMMARY,,Number of duplicate marked reads,123518125,11.23
    ...
    # then tumor and normal per-read-group metrics::
    TUMOR MAPPING/ALIGNING PER RG,T_SRR7890936_50pc,Total reads in RG,2200000000,100.00
    TUMOR MAPPING/ALIGNING PER RG,T_SRR7890936_50pc,Number of duplicate marked reads,433637413,19.71
    TUMOR MAPPING/ALIGNING PER RG,T_SRR7890936_50pc,Number of duplicate marked and mate reads removed,NA
    TUMOR MAPPING/ALIGNING PER RG,T_SRR7890936_50pc,Number of unique reads (excl. duplicate marked reads),1766362587,80.29
    TUMOR MAPPING/ALIGNING PER RG,T_SRR7890936_50pc,Reads with mate sequenced,2200000000,100.00
    TUMOR MAPPING/ALIGNING PER RG,T_SRR7890936_50pc,Reads without mate sequenced,0,0.00
    TUMOR MAPPING/ALIGNING PER RG,T_SRR7890936_50pc,QC-failed reads,0,0.00
    TUMOR MAPPING/ALIGNING PER RG,T_SRR7890936_50pc,Mapped reads,2130883930,96.86
    TUMOR MAPPING/ALIGNING PER RG,T_SRR7890936_50pc,Mapped reads R1,1066701794,96.97
    TUMOR MAPPING/ALIGNING PER RG,T_SRR7890936_50pc,Mapped reads R2,1064182136,96.74
    TUMOR MAPPING/ALIGNING PER RG,T_SRR7890936_50pc,Number of unique & mapped reads (excl. duplicate marked reads),1697246517,77.15
    TUMOR MAPPING/ALIGNING PER RG,T_SRR7890936_50pc,Unmapped reads,69116070,3.14
    TUMOR MAPPING/ALIGNING PER RG,T_SRR7890936_50pc,Singleton reads (itself mapped; mate unmapped),3917092,0.18
    TUMOR MAPPING/ALIGNING PER RG,T_SRR7890936_50pc,Paired reads (itself & mate mapped),2126966838,96.68
    TUMOR MAPPING/ALIGNING PER RG,T_SRR7890936_50pc,Properly paired reads,2103060370,95.59
    TUMOR MAPPING/ALIGNING PER RG,T_SRR7890936_50pc,Not properly paired reads (discordant),23906468,1.09
    TUMOR MAPPING/ALIGNING PER RG,T_SRR7890936_50pc,Paired reads mapped to different chromosomes,17454370,0.82
    TUMOR MAPPING/ALIGNING PER RG,T_SRR7890936_50pc,Paired reads mapped to different chromosomes (MAPQ>=10),6463547,0.30
    TUMOR MAPPING/ALIGNING PER RG,T_SRR7890936_50pc,Reads with MAPQ [40:inf),2002661377,91.03
    TUMOR MAPPING/ALIGNING PER RG,T_SRR7890936_50pc,Reads with MAPQ [30:40),7169392,0.33
    TUMOR MAPPING/ALIGNING PER RG,T_SRR7890936_50pc,Reads with MAPQ [20:30),16644390,0.76
    TUMOR MAPPING/ALIGNING PER RG,T_SRR7890936_50pc,Reads with MAPQ [10:20),20280057,0.92
    TUMOR MAPPING/ALIGNING PER RG,T_SRR7890936_50pc,Reads with MAPQ [ 0:10),84128714,3.82
    TUMOR MAPPING/ALIGNING PER RG,T_SRR7890936_50pc,Reads with MAPQ NA (Unmapped reads),69116070,3.14
    TUMOR MAPPING/ALIGNING PER RG,T_SRR7890936_50pc,Reads with indel R1,26849051,2.52
    TUMOR MAPPING/ALIGNING PER RG,T_SRR7890936_50pc,Reads with indel R2,24810803,2.33
    TUMOR MAPPING/ALIGNING PER RG,T_SRR7890936_50pc,Total bases,330000000000
    TUMOR MAPPING/ALIGNING PER RG,T_SRR7890936_50pc,Total bases R1,165000000000
    TUMOR MAPPING/ALIGNING PER RG,T_SRR7890936_50pc,Total bases R2,165000000000
    TUMOR MAPPING/ALIGNING PER RG,T_SRR7890936_50pc,Mapped bases R1,160005269100
    TUMOR MAPPING/ALIGNING PER RG,T_SRR7890936_50pc,Mapped bases R2,159627320400
    TUMOR MAPPING/ALIGNING PER RG,T_SRR7890936_50pc,Soft-clipped bases R1,1757128997,1.10
    TUMOR MAPPING/ALIGNING PER RG,T_SRR7890936_50pc,Soft-clipped bases R2,3208748350,2.01
    TUMOR MAPPING/ALIGNING PER RG,T_SRR7890936_50pc,Mismatched bases R1,585802788,0.37
    TUMOR MAPPING/ALIGNING PER RG,T_SRR7890936_50pc,Mismatched bases R2,1155805091,0.72
    TUMOR MAPPING/ALIGNING PER RG,T_SRR7890936_50pc,Mismatched bases R1 (excl. indels),501394281,0.31
    TUMOR MAPPING/ALIGNING PER RG,T_SRR7890936_50pc,Mismatched bases R2 (excl. indels),1073788605,0.67
    TUMOR MAPPING/ALIGNING PER RG,T_SRR7890936_50pc,Q30 bases,297564555927,90.17
    TUMOR MAPPING/ALIGNING PER RG,T_SRR7890936_50pc,Q30 bases R1,155492239719,94.24
    TUMOR MAPPING/ALIGNING PER RG,T_SRR7890936_50pc,Q30 bases R2,142072316208,86.10
    TUMOR MAPPING/ALIGNING PER RG,T_SRR7890936_50pc,Q30 bases (excl. dups & clipped bases),246555769158
    TUMOR MAPPING/ALIGNING PER RG,T_SRR7890936_50pc,Total alignments,2190085267
    TUMOR MAPPING/ALIGNING PER RG,T_SRR7890936_50pc,Secondary alignments,0
    TUMOR MAPPING/ALIGNING PER RG,T_SRR7890936_50pc,Supplementary (chimeric) alignments,59201337
    TUMOR MAPPING/ALIGNING PER RG,T_SRR7890936_50pc,Estimated read length,150.00
    TUMOR MAPPING/ALIGNING PER RG,T_SRR7890936_50pc,Average sequenced coverage over genome,102.83
    TUMOR MAPPING/ALIGNING PER RG,T_SRR7890936_50pc,Insert length: mean,383.01
    TUMOR MAPPING/ALIGNING PER RG,T_SRR7890936_50pc,Insert length: median,376.00
    TUMOR MAPPING/ALIGNING PER RG,T_SRR7890936_50pc,Insert length: standard deviation,87.58
    # same for normal:
    NORMAL MAPPING/ALIGNING PER RG,N_SRR7890889,Total reads in RG,1100000000,100.00
    NORMAL MAPPING/ALIGNING PER RG,N_SRR7890889,Number of duplicate marked reads,NA
    NORMAL MAPPING/ALIGNING PER RG,N_SRR7890889,Number of duplicate marked and mate reads removed,NA
    NORMAL MAPPING/ALIGNING PER RG,N_SRR7890889,Number of unique reads (excl. duplicate marked reads),NA
    ...

    We are reporting summary metrics in the general stats table, and per-read-group in a separate table.
    """

    f['s_name'] = re.search(r'(.*).mapping_metrics.csv', f['fn']).group(1)

    data_by_readgroup = defaultdict(dict)
    data_by_phenotype = defaultdict(dict)

    for line in f['f'].splitlines():
        fields = line.split(',')
        phenotype = fields[0].split('/')[0].split(' ')[0].lower()  # TUMOR MAPPING -> tumor
        analysis = fields[0].split('/')[1]  # ALIGNING SUMMARY, ALIGNING PER RG
        metric = fields[2]
        value = fields[3]
        try:
            value = int(value)
        except ValueError:
            try:
                value = float(value)
            except ValueError:
                pass

        percentage = None
        if len(fields) > 4:  # percentage
            percentage = fields[4]
            try:
                percentage = float(percentage)
            except ValueError:
                pass

        # sample-unspecific metrics are reported only in ALIGNING SUMMARY sections
        if analysis == 'ALIGNING SUMMARY':
            data_by_phenotype[phenotype][metric] = value
            if percentage is not None:
                data_by_phenotype[phenotype][metric + ' pct'] = percentage

        # for sample-specific metrics, using ALIGNING PER RG because it has the sample name in the 2nd col
        if analysis == 'ALIGNING PER RG':
            # setting normal and tumor sample names for future use
            readgroup = fields[1]
            data_by_readgroup[readgroup][metric] = value
            if percentage is not None:
                data_by_readgroup[readgroup][metric + ' pct'] = percentage

    # adding some missing values that we wanna report for consistency
    for data in itertools.chain(data_by_readgroup.values(), data_by_phenotype.values()):
        # fixing when deduplication wasn't performed
        if data['Number of duplicate marked reads'] == 'NA':
            data['Number of duplicate marked reads'] = 0
        if data['Number of duplicate marked and mate reads removed'] == 'NA':
            data['Number of duplicate marked and mate reads removed'] = 0
        if data['Number of unique reads (excl. duplicate marked reads)'] == 'NA':
            data['Number of unique reads (excl. duplicate marked reads)'] = data['Mapped reads']
        # adding alignment percentages
        if data['Total alignments'] > 0:
            data['Secondary alignments pct'] = data['Secondary alignments'] / data['Total alignments'] * 100.0
        # adding some missing bases percentages
        if data['Total bases'] > 0:
            data['Q30 bases (excl. dups & clipped bases) pct'] = data['Q30 bases (excl. dups & clipped bases)'] / data['Total bases'] * 100.0
            data['Mapped bases R1 pct'] = data['Mapped bases R1'] / data['Total bases'] * 100.0
            data['Mapped bases R2 pct'] = data['Mapped bases R2'] / data['Total bases'] * 100.0

    return data_by_readgroup, data_by_phenotype


MAPPING_METRICS = [
    # id_in_data                                                    # title (display name)  # in_genstats  # in_own_tabl # unit  # description
    # Read stats:
    Metric('Total input reads'                                      , 'Input reads'                , '#'  , None , 'reads', 'Total number of input reads for this sample (or total number of reads in all input read groups combined), {}'),
    Metric('Total reads in RG'                                      , 'Input reads'                , None , '#'  , 'reads', 'Total number of reads in this RG, {}'),
    Metric('Average sequenced coverage over genome'                 , 'Raw depth'                  , 'hid', '#'  , 'x'    , 'Average sequenced coverage over genome, based on all input reads (including duplicate, clipped and low quality bases and reads)'),
    Metric('Reads with mate sequenced'                              , 'Paired'                     , 'hid', '%'  , 'reads', 'Number of reads with a mate sequenced, {}'),
    Metric('Reads without mate sequenced'                           , 'Single'                     , None , 'hid', 'reads', 'Number of reads without a mate sequenced, {}', the_higher_the_worse=True),
    Metric('QC-failed reads'                                        , 'QC-fail'                    , 'hid', '%'  , 'reads', 'Number of reads not passing platform/vendor quality checks (SAM flag 0x200), {}', precision=2, the_higher_the_worse=True),
    Metric('Mapped reads'                                           , 'Map'                        , 'hid', 'hid', 'reads', 'Number of mapped reads, {}'),
    Metric('Mapped reads R1'                                        , 'Map R1'                     , None , 'hid', 'reads', 'Number of mapped reads R1, {}'),
    Metric('Mapped reads R2'                                        , 'Map R2'                     , None , 'hid', 'reads', 'Number of mapped reads R2, {}'),
    Metric('Unmapped reads'                                         , 'Unmap'                      , '%'  , '%'  , 'reads', 'Number of unmapped reads, {}', the_higher_the_worse=True),
    Metric('Reads with MAPQ [40:inf)'                               , 'MQ⩾40'                      , None , 'hid', 'reads', 'Number of reads with MAPQ [40:inf), {}'),
    Metric('Number of duplicate marked reads'                       , 'Dup'                        , '%'  , '%'  , 'reads', 'Number of duplicate marked reads as a result of the --enable-duplicatemarking option being used, {}', the_higher_the_worse=True),
    Metric('Number of duplicate marked and mate reads removed'      , 'Dup with mates removed'     , None , 'hid', 'reads', 'Number of reads marked as duplicates, along with any mate reads, that are removed '
                                                                                                                              'when the --remove-duplicates option is used, {}', the_higher_the_worse=True),
    Metric('Number of unique reads (excl. duplicate marked reads)'  , 'Uniq'                       , None , 'hid', 'reads', 'Number of unique reads (all reads minus duplicate marked reads), {}'),
    Metric('Number of unique & mapped reads (excl. duplicate marked reads)', 'Uniq map'            , 'hid', 'hid', 'reads', 'Number of unique & mapped reads (mapped reads minus duplicate marked reads), {}'),
    Metric('Paired reads (itself & mate mapped)'                    , 'Self & mate map'            , None , 'hid', 'reads', 'Number of reads mapped in pairs (itself & mate mapped), {}'),
    Metric('Properly paired reads'                                  , 'Prop pair'                  , '%'  , '%'  , 'reads', 'Number of properly paired reads, {} (both reads in pair are mapped and '
                                                                                                                              'fall within an acceptable range from each other based on the estimated insert length distribution). '
                                                                                                                              'Duplicate and low quality reads are not excluded, i.e. the percentage is calculated relative to the total number of reads.'),
    Metric('Not properly paired reads (discordant)'                 , 'Discord'                    , 'hid', '%'  , 'reads', 'Number of discordant reads: paired reads minus properly paired reads , {}', precision=2, the_higher_the_worse=True),
    Metric('Singleton reads (itself mapped; mate unmapped)'         , 'Singleton'                  , None , '%'  , 'reads', 'Number of singleton reads: itself mapped; mate unmapped, {}', precision=2, the_higher_the_worse=True),
    Metric('Paired reads mapped to different chromosomes'           , 'Diff chrom'                 , None , 'hid', 'reads', 'Number of paired reads with a mate mapped to a different chromosome, {}', precision=2, the_higher_the_worse=True),
    Metric('Paired reads mapped to different chromosomes (MAPQ>=10)', 'Diff chr, MQ⩾10'            , 'hid', '%'  , 'reads', 'Number of paired reads, mapped with MAPQ>=10 and with a mate mapped to a different chromosome, {}', precision=2, the_higher_the_worse=True),
    Metric('Reads with indel R1'                                    , 'Reads with indel R1'        , None , 'hid', 'reads', 'Number of R1 reads containing at least 1 indel, {}', the_higher_the_worse=True),
    Metric('Reads with indel R2'                                    , 'Reads with indel R2'        , None , 'hid', 'reads', 'Number of R2 reads containing at least 1 indel, {}', the_higher_the_worse=True),
    # Read length stats
    Metric('Estimated read length'                                  , 'Read len'                   , 'hid', 'hid', 'bp'   , 'Estimated read length. Total number of input bases divided by the number of reads'),
    Metric('Insert length: mean'                                    , 'Avg IS'                     , 'hid', 'hid', 'bp'   , 'Mean insert size'),
    Metric('Insert length: median'                                  , 'Med IS'                     , '#'  , '#'  , 'bp'   , 'Median insert size', precision=0),
    Metric('Insert length: standard deviation'                      , 'IS std'                     , 'hid', 'hid', 'bp'   , 'Standard deviation of insert size deviation'),
    # Bases stats:
    Metric('Total bases'                                            , 'Input bases'                , 'hid', 'hid', 'bases', 'Total number of bases sequenced, {}'),
    Metric('Total bases R1'                                         , 'Input bases R1'             , None , 'hid', 'bases', 'Total number of bases sequenced on R1 reads, {}'),
    Metric('Total bases R2'                                         , 'Input bases R2'             , None , 'hid', 'bases', 'Total number of bases sequenced on R2 reads, {}'),
    Metric('Mapped bases R1'                                        , 'Mapped bases R1'            , None , 'hid', 'bases', 'Number of mapped bases on R1 reads, {}'),
    Metric('Mapped bases R2'                                        , 'Mapped bases R2'            , None , 'hid', 'bases', 'Number of mapped bases on R2 reads, {}'),
    Metric('Soft-clipped bases R1'                                  , 'Soft-clip bases R1'         , None , 'hid', 'bases', 'Number of soft-clipped bases on R1 reads, {}', the_higher_the_worse=True),
    Metric('Soft-clipped bases R2'                                  , 'Soft-clip bases R2'         , None , 'hid', 'bases', 'Number of soft-clipped bases on R2 reads, {}', the_higher_the_worse=True),
    Metric('Mismatched bases R1'                                    , 'MM bases R1'                , None , 'hid', 'bases', 'Number of mismatched bases on R1, {}, which is the sum of SNP count and indel lengths. '
                                                                                                                              'It does not count anything within soft clipping, or RNA introns. '
                                                                                                                              'It also does not count a mismatch if either the reference base or read base is N', the_higher_the_worse=True),
    Metric('Mismatched bases R2'                                    , 'MM bases R2'                , None , 'hid', 'bases', 'Number of mismatched bases on R2, {}, which is the sum of SNP count and indel lengths. '
                                                                                                                              'It does not count anything within soft clipping, or RNA introns. '
                                                                                                                              'It also does not count a mismatch if either the reference base or read base is N', the_higher_the_worse=True),
    Metric('Mismatched bases R1 (excl. indels)'                     , 'MM bases R1 excl indels'    , None , 'hid', 'bases', 'Number of mismatched bases on R1, {}. The indels lengts are ignored. '
                                                                                                                              'It does not count anything within soft clipping, or RNA introns. '
                                                                                                                              'It also does not count a mismatch if either the reference base or read base is N', the_higher_the_worse=True),
    Metric('Mismatched bases R2 (excl. indels)'                     , 'MM bases R2 excl indels'    , None , 'hid', 'bases', 'Number of mismatched bases on R2, {}. The indels lengts are ignored. '
                                                                                                                              'It does not count anything within soft clipping, or RNA introns. '
                                                                                                                              'It also does not count a mismatch if either the reference base or read base is N', the_higher_the_worse=True),
    Metric('Q30 bases'                                              , 'Q30'                        , 'hid', 'hid', 'bases', 'Number of raw bases with BQ >= 30, {}'),
    Metric('Q30 bases R1'                                           , 'Q30 R1'                     , None , 'hid', 'bases', 'Number of raw bases on R1 reads with BQ >= 30, {}'),
    Metric('Q30 bases R2'                                           , 'Q30 R2'                     , None , 'hid', 'bases', 'Number of raw bases on R2 reads with BQ >= 30, {}'),
    Metric('Q30 bases (excl. dups & clipped bases)'                 , 'Q30 excl dup & clipped'     , None , 'hid', 'bases', 'Number of non-clipped bases with BQ >= 30 on non-duplicate reads, {}'),
    # Alignments stats:
    Metric('Total alignments'                                       , 'Alignments'                 , 'hid', '#'  , 'reads', 'Total number of alignments with MQ > 0, {}'),
    Metric('Secondary alignments'                                   , 'Sec\'ry'                    , 'hid', '%'  , 'reads', 'Number of secondary alignments, {}. Secondary alignment occurs when '
                                                                                                                              'a given read could align reasonably well to more than one place. '
                                                                                                                              'One of the possible reported alignments is termed "primary" and '
                                                                                                                              'the others will be marked as "secondary".', precision=2),
    Metric('Supplementary (chimeric) alignments'                    , 'Suppl\'ry'                  , 'hid', '%'  , 'reads', 'Number of supplementary (chimeric) alignments, {}. A chimeric read is split '
                                                                                                                              'over multiple loci (possibly due to structural variants). One alignment is '
                                                                                                                              'referred to as the representative alignment, the other are supplementary', precision=2),
]



#!/usr/bin/env python

""" MultiQC plugin hooks. Enables MultiQC plugins
to run their own custom subroutines at predefined
trigger points during MultiQC execution. """

import pkg_resources

# Load the hooks
hook_functions = {}
for entry_point in pkg_resources.iter_entry_points('multiqc.hooks.v1'):
  nicename = str(entry_point).split('=')[0].strip()
  try:
    hook_functions[nicename].append(entry_point.load())
  except KeyError:
    hook_functions[nicename] = [entry_point.load()]

# Function to run the hooks
def mqc_trigger (trigger):
  for hook in hook_functions.get(trigger, []):
    hook()



#!/usr/bin/env python

""" MultiQC module to parse output from Bowtie 1 """

from __future__ import print_function
from collections import OrderedDict
import logging
import re

from multiqc import config
from multiqc.plots import bargraph
from multiqc.modules.base_module import BaseMultiqcModule

# Initialise the logger
log = logging.getLogger(__name__)

class MultiqcModule(BaseMultiqcModule):
    """ Bowtie 1 module, parses stderr logs. """

    def __init__(self):

        # Initialise the parent object
        super(MultiqcModule, self).__init__(name='Bowtie 1', anchor='bowtie1',
        target='Bowtie 1', href="http://bowtie-bio.sourceforge.net/",
        info="is an ultrafast, memory-efficient short read aligner.")

        # Find and load any Bowtie reports
        self.bowtie_data = dict()
        for f in self.find_log_files('bowtie1'):
            self.parse_bowtie_logs(f)

        # Filter to strip out ignored sample names
        self.bowtie_data = self.ignore_samples(self.bowtie_data)

        if len(self.bowtie_data) == 0:
            raise UserWarning

        log.info("Found {} reports".format(len(self.bowtie_data)))

        # Write parsed report data to a file
        self.write_data_file(self.bowtie_data, 'multiqc_bowtie1')

        # Basic Stats Table
        # Report table is immutable, so just updating it works
        self.bowtie_general_stats_table()

        # Alignment Rate Plot
        self.bowtie_alignment_plot()


    def parse_bowtie_logs(self, f):
        s_name = f['s_name']
        parsed_data = {}
        regexes = {
            'reads_processed': r"# reads processed:\s+(\d+)",
            'reads_aligned': r"# reads with at least one reported alignment:\s+(\d+)",
            'reads_aligned_percentage': r"# reads with at least one reported alignment:\s+\d+\s+\(([\d\.]+)%\)",
            'not_aligned': r"# reads that failed to align:\s+(\d+)",
            'not_aligned_percentage': r"# reads that failed to align:\s+\d+\s+\(([\d\.]+)%\)",
            'multimapped': r"# reads with alignments suppressed due to -m:\s+(\d+)",
            'multimapped_percentage': r"# reads with alignments suppressed due to -m:\s+\d+\s+\(([\d\.]+)%\)"
        }
        for l in f['f'].splitlines():
            # Attempt in vain to find original bowtie1 command, logged by another program
            if 'bowtie' in l and 'q.gz' in l:
                fqmatch = re.search(r"([^\s,]+\.f(ast)?q.gz)", l)
                if fqmatch:
                    s_name = self.clean_s_name(fqmatch.group(1), f['root'])
                    log.debug("Found a bowtie command, updating sample name to '{}'".format(s_name))

            # End of log, reset in case there is another in this file
            if 'Overall time:' in l:
                if len(parsed_data) > 0:
                    if s_name in self.bowtie_data:
                        log.debug("Duplicate sample name found! Overwriting: {}".format(s_name))
                    self.add_data_source(f, s_name)
                    self.bowtie_data[s_name] = parsed_data
                s_name = f['s_name']
                parsed_data = {}

            for k, r in regexes.items():
                match = re.search(r, l)
                if match:
                    parsed_data[k] = float(match.group(1).replace(',', ''))

        if len(parsed_data) > 0:
            if s_name in self.bowtie_data:
                log.debug("Duplicate sample name found! Overwriting: {}".format(s_name))
            self.add_data_source(f, s_name)
            self.bowtie_data[s_name] = parsed_data


    def bowtie_general_stats_table(self):
        """ Take the parsed stats from the Bowtie report and add it to the
        basic stats table at the top of the report """

        headers = OrderedDict()
        headers['reads_aligned_percentage'] = {
            'title': '% Aligned',
            'description': '% reads with at least one reported alignment',
            'max': 100,
            'min': 0,
            'suffix': '%',
            'scale': 'YlGn'
        }
        headers['reads_aligned'] = {
            'title': '{} Aligned'.format(config.read_count_prefix),
            'description': 'reads with at least one reported alignment ({})'.format(config.read_count_desc),
            'min': 0,
            'scale': 'PuRd',
            'modify': lambda x: x * config.read_count_multiplier,
            'shared_key': 'read_count'
        }
        self.general_stats_addcols(self.bowtie_data, headers)

    def bowtie_alignment_plot (self):
        """ Make the HighCharts HTML to plot the alignment rates """

        # Specify the order of the different possible categories
        keys = OrderedDict()
        keys['reads_aligned'] = { 'color': '#8bbc21', 'name': 'Aligned' }
        keys['multimapped'] =   { 'color': '#2f7ed8', 'name': 'Multimapped' }
        keys['not_aligned'] =   { 'color': '#0d233a', 'name': 'Not aligned' }

        # Config for the plot
        config = {
            'id': 'bowtie1_alignment',
            'title': 'Bowtie 1: Alignment Scores',
            'ylab': '# Reads',
            'cpswitch_counts_label': 'Number of Reads'
        }

        self.add_section(
            description = 'This plot shows the number of reads aligning to the reference in different ways.',
            helptext = '''
            There are 3 possible types of alignment:
            * **Aligned**: Read has only one occurence in the reference genome.
            * **Multimapped**: Read has multiple occurence.
            * **Not aligned**: Read has no occurence.
            ''',
            plot = bargraph.plot(self.bowtie_data, keys, config)
        )



#!/usr/bin/env python
# -*- coding: utf-8 -*-

""" MultiQC submodule to parse output from Samtools idxstats """

import logging
from collections import OrderedDict, defaultdict
from multiqc import config
from multiqc.plots import bargraph, linegraph

# Initialise the logger
log = logging.getLogger(__name__)

class IdxstatsReportMixin():

    def parse_samtools_idxstats(self):
        """ Find Samtools idxstats logs and parse their data """

        self.samtools_idxstats = dict()
        for f in self.find_log_files('samtools/idxstats'):
            parsed_data = parse_single_report(f['f'])
            if len(parsed_data) > 0:
                if f['s_name'] in self.samtools_idxstats:
                    log.debug("Duplicate sample name found! Overwriting: {}".format(f['s_name']))
                self.add_data_source(f, section='idxstats')
                self.samtools_idxstats[f['s_name']] = parsed_data

        # Filter to strip out ignored sample names
        self.samtools_idxstats = self.ignore_samples(self.samtools_idxstats)

        if len(self.samtools_idxstats) > 0:

            # Write parsed report data to a file (restructure first)
            self.write_data_file(self.samtools_idxstats, 'multiqc_samtools_idxstats')

            # Prep the data for the plots
            keys = list()
            pdata = dict()
            pdata_norm = dict()
            pdata_obs_exp = dict()
            xy_counts = dict()
            # Count the total mapped reads for every chromosome
            chrs_mapped = defaultdict(lambda:0)
            sample_mapped = defaultdict(lambda:0)
            total_mapped = 0
            # Cutoff, can be customised in config
            cutoff = float(getattr(config, 'samtools_idxstats_fraction_cutoff', 0.001))
            if cutoff != 0.001:
                log.info('Setting idxstats cutoff to: {}%'.format(cutoff*100.0))
            for s_name in self.samtools_idxstats:
                for chrom in self.samtools_idxstats[s_name]:
                    chrs_mapped[chrom] += self.samtools_idxstats[s_name][chrom][0]
                    sample_mapped[s_name] += self.samtools_idxstats[s_name][chrom][0]
                    total_mapped += self.samtools_idxstats[s_name][chrom][0]
            req_reads = float(total_mapped)*cutoff
            chr_always = getattr(config, 'samtools_idxstats_always', [])
            if len(chr_always) > 0:
                log.info('Trying to include these chromosomes in idxstats: {}'.format(', '.join(chr_always)))
            chr_ignore = getattr(config, 'samtools_idxstats_ignore', [])
            if len(chr_ignore) > 0:
                log.info('Excluding these chromosomes from idxstats: {}'.format(', '.join(chr_ignore)))
            xchr = getattr(config, 'samtools_idxstats_xchr', False)
            if xchr:
                log.info('Using "{}" as X chromosome name'.format(xchr))
            ychr = getattr(config, 'samtools_idxstats_ychr', False)
            if ychr:
                log.info('Using "{}" as Y chromosome name'.format(ychr))
            # Go through again and collect all of the keys that have enough counts
            # Also get the X/Y counts if we find them
            for s_name in self.samtools_idxstats:
                x_count = False
                y_count = False
                for chrom in self.samtools_idxstats[s_name]:
                    if float(chrs_mapped[chrom]) > req_reads or chrom in chr_always:
                        if chrom not in chr_ignore and chrom not in keys:
                            keys.append(chrom)
                    # Collect X and Y counts if we have them
                    mapped = self.samtools_idxstats[s_name][chrom][0]
                    if xchr is not False :
                        if str(xchr) == str(chrom):
                            x_count = mapped
                    else:
                        if chrom.lower() == 'x' or chrom.lower() == 'chrx':
                            x_count = mapped
                    if ychr is not False :
                        if str(ychr) == str(chrom):
                            y_count = mapped
                    else:
                        if chrom.lower() == 'y' or chrom.lower() == 'chry':
                            y_count = mapped
                # Only save these counts if we have both x and y
                if x_count and y_count:
                    xy_counts[s_name] = {'x': x_count, 'y': y_count }
            # Ok, one last time. We have the chromosomes that we want to plot,
            # now collect the counts
            for s_name in self.samtools_idxstats:
                pdata[s_name] = OrderedDict()
                pdata_norm[s_name] = OrderedDict()
                pdata_obs_exp[s_name] = OrderedDict()
                genome_size = float(sum([stats[1] for stats in self.samtools_idxstats[s_name].values()]))
                for k in keys:

                    try:
                        pdata[s_name][k] = self.samtools_idxstats[s_name][k][0]
                        pdata_norm[s_name][k] = float(self.samtools_idxstats[s_name][k][0]) / sample_mapped[s_name]
                        chrom_size = float(self.samtools_idxstats[s_name][k][1])
                        expected_count = (chrom_size / genome_size) * float(sample_mapped[s_name])
                        pdata_obs_exp[s_name][k] = float(pdata[s_name][k]) / expected_count
                    except (KeyError, ZeroDivisionError):
                        pdata[s_name][k] = 0
                        pdata_norm[s_name][k] = 0
                        pdata_obs_exp[s_name][k] = 0

            # X/Y ratio plot
            if len(xy_counts) > 0:
                xy_keys = OrderedDict()
                xy_keys['x'] = { 'name': xchr if xchr else 'Chromosome X' }
                xy_keys['y'] = { 'name': ychr if ychr else 'Chromosome Y' }
                pconfig = {
                    'id': 'samtools-idxstats-xy-plot',
                    'title': 'Samtools idxstats: chrXY mapped reads',
                    'ylab': 'Percent of X+Y Reads',
                    'cpswitch_counts_label': 'Number of Reads',
                    'cpswitch_percent_label': 'Percent of X+Y Reads',
                    'cpswitch_c_active': False
                }
                self.add_section (
                    name = 'XY counts',
                    anchor = 'samtools-idxstats-xy-counts',
                    plot = bargraph.plot(xy_counts, xy_keys, pconfig)
                )


            # Mapped reads per chr line plot
            pconfig = {
                'id': 'samtools-idxstats-mapped-reads-plot',
                'title': 'Samtools idxstats: Mapped reads per contig',
                'ylab': '# mapped reads',
                'xlab': 'Chromosome Name',
                'logswitch': True,
                'categories': True,
                'tt_label': '<strong>{point.category}:</strong> {point.y:.2f}',
                'data_labels': [
                    {'name': 'Normalised Counts', 'ylab': 'Fraction of total count'},
                    {'name': 'Observed over Expected Counts', 'ylab': 'log10 ( Observed over expected counts )'},
                    {'name': 'Raw Counts', 'ylab': '# mapped reads'}
                ]
            }
            self.add_section (
                name = 'Mapped reads per contig',
                anchor = 'samtools-idxstats',
                description = 'The <code>samtools idxstats</code> tool counts the number of mapped reads per chromosome / contig. ' +
                    'Chromosomes with &lt; {}% of the total aligned reads are omitted from this plot.'.format(cutoff*100),
                plot = linegraph.plot([pdata_norm, pdata_obs_exp, pdata], pconfig)
            )

        # Return the number of logs that were found
        return len(self.samtools_idxstats)


# idxstats has four columns: chr, length, mapped, unmapped
# http://www.htslib.org/doc/samtools.html

def parse_single_report(f):
    """ Parse a samtools idxstats idxstats """

    parsed_data = OrderedDict()
    for l in f.splitlines():
        s = l.split("\t")
        try:
            parsed_data[s[0]] = [int(s[2]), int(s[1])]
        except (IndexError, ValueError):
            pass
    return parsed_data



from itertools import chain

from multiqc.plots import linegraph

def plot_bqhist(samples, file_type, **plot_args):
    """ Create line graph plot of histogram data for BBMap 'bqhist' output.

    The 'samples' parameter could be from the bbmap mod_data dictionary:
    samples = bbmap.MultiqcModule.mod_data[file_type]
    """

    all_x = set()
    for item in sorted(chain(*[samples[sample]['data'].items()
                                for sample in samples])):
        all_x.add(item[0])


    columns_to_plot = {
        #'Counts': {
        #    0: 'Read 1',
        #    9: 'Read 2',
        #},
        'Read 1 averages': {
            #1: 'Min',
            #2: 'Max',
            3: 'Mean',
            #4: '1st quartile',
            #5: 'Median',
            #6: '3rd quartile',
            #7: 'Left (lower) whisker',
            #8: 'Right (upper) whisker',
        },
        'Read 2 averages': {
            #10: 'Min',
            #11: 'Max',
            12: 'Mean',
            #13: '1st quartile',
            #14: 'Median',
            #15: '3rd quartile',
            #16: 'Left (lower) whisker',
            #17: 'Right (upper) whisker',
        },
    }

    plot_data = []
    for column_type in columns_to_plot:
        plot_data.append(
            {
                sample+'.'+column_name: {
                    x: samples[sample]['data'][x][column] if x in samples[sample]['data'] else 0
                    for x in all_x
                }
                for sample in samples
                for column, column_name in columns_to_plot[column_type].items()
            }
        )

    plot_params = {
            'id': 'bbmap-' + file_type + '_plot',
            'title': 'BBTools: ' + plot_args['plot_title'],
            'xlab': 'Read position',
            'ylab': 'Average quality score',
            'data_labels': [
                #{'name': 'Count histogram', 'ylab': 'Read count'},
                {'name': 'Read 1',},
                {'name': 'Read 2',},
            ]
    }

    plot_params.update(plot_args['plot_params'])
    plot = linegraph.plot(
        plot_data,
        plot_params
    )

    return plot
